{% extends "base.html" %}
{% block title %}Help / Docs / Code - BiweeklyBudget{% endblock %}
{% block body %}
{% include 'notifications.html' %}
            <!-- /.row -->
            <div class="row" id="content-row">
                <div class="col-lg-12">
                    <ul>
                        <li><strong>Project Homepage:</strong> <a href="{{ url }}">{{ url }}</a></li>
                        <li><strong>Documentation:</strong> <a href="http://biweeklybudget.readthedocs.io/">http://biweeklybudget.readthedocs.io/</a> <em>(for this version: <a href="http://biweeklybudget.readthedocs.io/en/{{ version }}/">http://biweeklybudget.readthedocs.io/en/{{ version }}/</a>)</em></li>
                        <li><strong>Issues / Bug Reports / Support:</strong> <a href="https://github.com/jantman/biweeklybudget/issues">https://github.com/jantman/biweeklybudget/issues</a></li>
                        <li><strong>Code Repository:</strong> <a href="https://github.com/jantman/biweeklybudget">https://github.com/jantman/biweeklybudget</a></li>
                    </ul>
                    <p><strong>biweeklybudget is <a href="https://www.gnu.org/licenses/agpl-3.0.en.html">AGPL 3.0</a> Free/Open Source Software. Anyone who can access or interact with this interface is entitled to the full source code of this exact running version.</strong></p>
                </div>
                <!-- /.col-lg-12 -->
                <div class="col-lg-12">
                    <div class="panel panel-default">
                        <div class="panel-heading">Support Information</div>
                        <!-- /.panel-heading -->
                        <div class="panel-body">
                            <dl>
                                <dt>Database URI</dt>
                                <dd>{{ db_uri }}</dd>
                                <dt>Version</dt>
                                <dd>{{ ver_info }}</dd>
                                <dt>User-Agent</dt>
                                <dd>{{ ua_str }}</dd>
                            </dl>
                        </div>
                        <!-- /.panel-body -->
                    </div>
                    <!-- /.panel-default
                </div>
                <!-- /.col-lg-12 -->
            </div>
            <!-- /.row -->
{% endblock %}
