"""add ScheduledTransaction sales_tax field

Revision ID: 1bb9e6a1c07c
Revises: d01774fa3ae3
Create Date: 2026-01-10 13:34:45.628528

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1bb9e6a1c07c'
down_revision = 'd01774fa3ae3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'scheduled_transactions',
        sa.Column('sales_tax', sa.Numeric(precision=10, scale=4), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('scheduled_transactions', 'sales_tax')
    # ### end Alembic commands ###
