"""add Transaction field to store sales tax paid

Revision ID: d01774fa3ae3
Revises: f5a002127934
Create Date: 2024-01-24 19:47:44.629591

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd01774fa3ae3'
down_revision = 'f5a002127934'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'transactions',
        sa.Column('sales_tax', sa.Numeric(precision=10, scale=4), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('transactions', 'sales_tax')
    # ### end Alembic commands ###
