<!DOCTYPE html>
<html lang="en">

<head>

    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="description" content="">
    <meta name="author" content="">

    <title>{% block title %}BiweeklyBudget{% endblock %}</title>

    <!-- Bootstrap Core CSS -->
    <link href="/static/startbootstrap-sb-admin-2/vendor/bootstrap/css/bootstrap.min.css" rel="stylesheet">

    <!-- MetisMenu CSS -->
    <link href="/static/startbootstrap-sb-admin-2/vendor/metisMenu/metisMenu.min.css" rel="stylesheet">

    <!-- Custom CSS -->
    <link href="/static/startbootstrap-sb-admin-2/dist/css/sb-admin-2.css" rel="stylesheet">

    <!-- Custom Fonts -->
    <link href="/static/startbootstrap-sb-admin-2/vendor/font-awesome/css/font-awesome.min.css" rel="stylesheet" type="text/css">

    <!-- App custom CSS -->
    <link href="/static/css/custom.css" rel="stylesheet" type="text/css">

    <!-- Python settings we need to be able to access from Javascript -->
    <script>
        var CURRENCY_SYMBOL = "{{ CURRENCY_SYM }}";
        var LOCALE_NAME = "{{ settings.LOCALE_NAME }}";
        if(LOCALE_NAME.indexOf('.') !== -1) {
          LOCALE_NAME = LOCALE_NAME.substr(0, LOCALE_NAME.indexOf('.'));
        }
        if(LOCALE_NAME.indexOf('_') !== -1) { LOCALE_NAME = LOCALE_NAME.replace('_', '-'); }
        var CURRENCY_CODE = "{{ settings.CURRENCY_CODE }}";
        var FUEL_VOLUME_UNIT = "{{ settings.FUEL_VOLUME_UNIT }}";
        var FUEL_VOLUME_ABBREVIATION = "{{ settings.FUEL_VOLUME_ABBREVIATION }}";
        var FUEL_ECO_ABBREVIATION = "{{ settings.FUEL_ECO_ABBREVIATION }}";
        var DISTANCE_UNIT = "{{ settings.DISTANCE_UNIT }}";
        var DISTANCE_UNIT_ABBREVIATION = "{{ settings.DISTANCE_UNIT_ABBREVIATION }}";
    </script>

    <!-- apply custom JS -->
    <script src="/static/js/custom.js"></script>

    {% block extra_head_css %}{% endblock %}

    <!-- HTML5 Shim and Respond.js IE8 support of HTML5 elements and media queries -->
    <!-- WARNING: Respond.js doesn't work if you view the page via file:// -->
    <!--[if lt IE 9]>
        <script src="https://oss.maxcdn.com/libs/html5shiv/3.7.0/html5shiv.js"></script>
        <script src="https://oss.maxcdn.com/libs/respond.js/1.4.2/respond.min.js"></script>
    <![endif]-->

</head>
<body>

    <div id="wrapper">

        <!-- Navigation -->
        <nav class="navbar navbar-default navbar-static-top" role="navigation" style="margin-bottom: 0">
            <div class="navbar-header">
                <button type="button" class="navbar-toggle" data-toggle="collapse" data-target=".navbar-collapse">
                    <span class="sr-only">Toggle navigation</span>
                    <span class="icon-bar"></span>
                    <span class="icon-bar"></span>
                    <span class="icon-bar"></span>
                </button>
                <a class="navbar-brand" href="index.html">{{ self.title() }}</a>
            </div>
            <!-- /.navbar-header -->

{% include 'nav.html' %}

        <div id="page-wrapper">
{% block body %}{% endblock %}
        </div>
        <!-- /#page-wrapper -->

    </div>
    <!-- /#wrapper -->

    <!-- jQuery -->
    <script src="/static/startbootstrap-sb-admin-2/vendor/jquery/jquery.min.js"></script>

    <!-- Bootstrap Core JavaScript -->
    <script src="/static/startbootstrap-sb-admin-2/vendor/bootstrap/js/bootstrap.min.js"></script>

    <!-- Metis Menu Plugin JavaScript -->
    <script src="/static/startbootstrap-sb-admin-2/vendor/metisMenu/metisMenu.min.js"></script>

    <!-- Custom Theme JavaScript -->
    <script src="/static/startbootstrap-sb-admin-2/dist/js/sb-admin-2.js"></script>
    {% block extra_foot_script %}{% endblock %}
</body>

</html>
