{% extends "base.html" %}
{% block title %}Example Template - BiweeklyBudget{% endblock %}
{% block body %}
{% include 'notifications.html' %}
            <div class="row" id="content-row">
                <div class="col-lg-12">
                    <h1 class="page-header">Example Template</h1>
                </div>
                <!-- /.col-lg-12 -->
            </div>
            <!-- /.row -->
{% endblock %}