        <!-- BEGIN nav.html -->
            <div class="navbar-default sidebar" role="navigation">
                <div class="sidebar-nav navbar">
                    <ul class="nav" id="side-menu">
                        <li>
                            <a href="/"><i class="fa fa-home fa-fw"></i> Home</a>
                        </li>
                        <li>
                            <a href="/payperiods"><i class="fa fa-calendar fa-fw"></i> Pay Periods</a>
                        </li>
                        <li>
                            <a href="/accounts"><i class="fa fa-bank fa-fw"></i> Accounts</a>
                        </li>
                        <li>
                            <a href="/accounts/credit-payoff"><i class="fa fa-credit-card fa-fw"></i> Credit Payoffs</a>
                        </li>
                        <li>
                            <a href="/ofx"><i class="fa fa-exchange fa-fw"></i> OFX</a>
                        </li>
                        <li>
                            <a href="/plaid-update"><i class="fa fa-file-code-o fa-fw"></i> Plaid Update</a>
                        </li>
                        <li>
                            <a href="/transactions"><i class="fa fa-money fa-fw"></i> Transactions</a>
                        </li>
                        <li>
                            <a href="/reconcile"><i class="fa fa-check fa-fw"></i> Reconcile</a>
                        </li>
                        <li>
                            <a href="/budgets"><i class="fa fa-envelope fa-fw"></i> Budgets</a>
                        </li>
                        <li>
                            <a href="/scheduled"><i class="fa fa-clock-o fa-fw"></i> Scheduled</a>
                        </li>
                        <li>
                            <a href="/fuel"><i class="fa fa-car fa-fw"></i> Fuel Log</a>
                        </li>
                        <li>
                            <a href="/projects"><i class="fa fa-th-list fa-fw"></i> Projects / BoM</a>
                        </li>
                        <li>&nbsp;</li>
                        <li>
                            <a href="/help"><i class="fa fa-support fa-fw"></i> Help/Docs/Code (AGPL)</a>
                        </li>
                    </ul>
                </div>
                <!-- /.sidebar-collapse -->
            </div>
            <!-- /.navbar-static-side -->
        </nav>
        <!-- END nav.html -->