"""add plaid_item_id and plaid_token attributes on Account

Revision ID: ac93b5c826a5
Revises: 073142f641b3
Create Date: 2020-01-26 20:57:22.045669

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ac93b5c826a5'
down_revision = '073142f641b3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'accounts',
        sa.Column('plaid_item_id', sa.String(length=70), nullable=True)
    )
    op.add_column(
        'accounts',
        sa.Column('plaid_token', sa.String(length=70), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('accounts', 'plaid_token')
    op.drop_column('accounts', 'plaid_item_id')
    # ### end Alembic commands ###
