span.data_age {
    font-size: .7em;
    font-style: italic;
    font-color: #333;
}

tr.inactive {
    background-color: #d9d9d9 !important;
}

div.reconcile-drop-target {
  min-height: 10px;
}

div.reconcile {
  box-sizing: border-box;
  min-height: 20px;
  padding-top: 9px;
  padding-right: 9px;
  padding-bottom: 9px;
  padding-left: 9px;
  margin-bottom: 20px;
  background-color: rgb(245, 245, 245);
  border-top-width: 1px;
  border-right-width: 1px;
  border-bottom-width: 1px;
  border-left-width: 1px;
  border-top-style: solid;
  border-right-style: solid;
  border-bottom-style: solid;
  border-left-style: solid;
  border-top-color: rgb(227, 227, 227);
  border-right-color: rgb(227, 227, 227);
  border-bottom-color: rgb(227, 227, 227);
  border-left-color: rgb(227, 227, 227);
  border-image-source: initial;
  border-image-slice: initial;
  border-image-width: initial;
  border-image-outset: initial;
  border-image-repeat: initial;
  border-top-left-radius: 3px;
  border-top-right-radius: 3px;
  border-bottom-right-radius: 3px;
  border-bottom-left-radius: 3px;
  box-shadow: rgba(0, 0, 0, 0.05) 0px 1px 1px inset;
}

div.reconcile-ofx-dropped {
  margin-bottom: 10px;
  margin-top: 10px;
  border-top-color: rgb(188, 232, 241);
  border-right-color: rgb(188, 232, 241);
  border-bottom-color: rgb(188, 232, 241);
  border-left-color: rgb(188, 232, 241);
  background-color: #d9edf7;

}

.disabledEditLink {
  font-weight: bold;
}

/*
height of .panel-vert-scroll is based on:
  Top header is 51px
  Panel heading is 41px
  Bottom padding on the txn/OFX panel is 20px
  Bottom submit row is 86px
  Add 2px for buffer to prevent body scrolling
  72px for the status bar row (added via JS)
----------
sum: 272px
----------

this was based on bootstrap3's .pre-scrollable
*/

.panel-vert-scroll {
  min-height: 100px;
  max-height: calc(100vh - 272px);
  overflow-y: scroll;
}
